// ctwain.cpp - implementation of application i/f to TWAIN

#include "stdafx.h"
#include "ctwain.h"

#ifdef _WIN32
  #define DSM_FILENAME "TWAIN_32.DLL"
  #define DSM_ENTRYPOINT "DSM_Entry"
#else
  #define DSM_FILENAME "TWAIN.DLL"
  #define DSM_ENTRYPOINT "DSM_ENTRY"
#endif

#define IsValidHandle(h) ((h) != NULL)

#define TRACE_ERROR(e) { if (bTrace) TRACE("TWAIN: Error(%d)\n", e); TwainError(e); }

static int			iAvailable;			// TWAIN available: 0:unknown, -1:No, 1:Yes
static DSMENTRYPROC	pSM_Entry;			// entry point of Data Source Manager (TWAIN.DLL)
static HINSTANCE	hSMLib;			// handle of SM

// default application identity structure:
const TW_IDENTITY	DefAppId = {
	0,									// Id, filled in by SM
	{ 1, 0, TWLG_USA, TWCY_USA, "<?>"},	// Version
    TWON_PROTOCOLMAJOR,
    TWON_PROTOCOLMINOR,
    DG_IMAGE | DG_CONTROL,
    "<?>",								// Mfg
    "<?>",								// Family
    "<?>"								// Product
    };
static TW_IDENTITY	SourceId;			// source identity structure




// CTwain class members
CTwain::CTwain(void)
 : pDefWnd(NULL), nState(NO_TWAIN_STATE)
{
	bShowUI = TRUE;
	hDib = NULL;
	rc = TWRC_SUCCESS;
	// turn off if you get tired of seeing the output:
	bTrace = TRUE;

	AppId = DefAppId;			// provide default registration
	SetState(PRE_SESSION);
}


void CTwain::RegisterApp(	// record application information
	int		nMajorNum, int nMinorNum,	// major and incremental revision of application. E.g.
										// for version 2.1, nMajorNum == 2 and nMinorNum == 1
	int		nLanguage,					// language of this version (use TWLG_xxx from TWAIN.H)
	int		nCountry,					// country of this version (use TWCY_xxx from TWAIN.H)
	LPSTR	lpszVersion,				// version info string e.g. "1.0b3 Beta release"
	LPSTR	lpszMfg,					// name of manufacturer/developer e.g. "Crazbat Software"
	LPSTR	lpszFamily,					// product family e.g. "BitStomper"
	LPSTR	lpszProduct)				// specific product e.g. "BitStomper Deluxe Pro"
{
    AppId.Id = 0;						// init to 0, but Source Manager will assign real value
    AppId.Version.MajorNum = nMajorNum;
    AppId.Version.MinorNum = nMinorNum;
    AppId.Version.Language = nLanguage;
    AppId.Version.Country  = nCountry;
    lstrcpy (AppId.Version.Info,  lpszVersion);

    AppId.ProtocolMajor =    TWON_PROTOCOLMAJOR;
    AppId.ProtocolMinor =    TWON_PROTOCOLMINOR;
    AppId.SupportedGroups =  DG_IMAGE | DG_CONTROL;
    lstrcpy (AppId.Manufacturer,  lpszMfg);
    lstrcpy (AppId.ProductFamily, lpszFamily);
    lstrcpy (AppId.ProductName,   lpszProduct);

}

CTwain::~CTwain()
{
	// shut down the Twain connection
	CloseSourceManager();
	UnloadSourceManager();
}

int CTwain::TwainAvailable(void)
{
	if (pSM_Entry) return TRUE;		// SM currently loaded

	if (iAvailable == 0) {
		if (LoadSourceManager()) {
			iAvailable = 1;
		} else {
			iAvailable = -1;
		}
	}
	return (iAvailable > 0);
}


TW_STATE CTwain::State(void)
{
	return nState;
}


void CTwain::SetState(TW_STATE nS)
{
#ifdef _DEBUG
	if (bTrace) {
		const char *pzState[] = {
			"",
			"1:PRE_SESSION",
			"2:SOURCE_MANAGER_LOADED",
			"3:SOURCE_MANAGER_OPEN",
			"4:SOURCE_OPEN",
			"5:SOURCE_ENABLED",
			"6:TRANSFER_READY",
			"7:TRANSFERRING"
		};
		TRACE("TWAIN: State %s -> %s\n", pzState[nState], pzState[nS]);
	}
#endif

	nState = nS;				// update the 'global'
	StateChange(nS);			// notify derived classes via callback
} // SetState


void CTwain::SetDefWindow(CWnd* pWnd)
// Sets the CWnd* to be used as the default for all other calls that
// take a window argument - SelectSource, OpenSourceManager, etc.
// If no default window is set, AfxGetMainWnd() is used.
{
	pDefWnd = pWnd;
}


CWnd* CTwain::DefWnd(CWnd* pWnd)
// Returns pWnd if it's non-null, otherwise it
// finds a safe non-null default substitute.
{
	return pWnd ? pWnd : pDefWnd ? pDefWnd : AfxGetMainWnd();
} // DefWnd


int CTwain::SelectSource(CWnd* pWnd)
{
	TW_IDENTITY		NewSourceId;
	TW_STATE		nStartState = State();
	int				bSuccess = FALSE;

	pWnd = DefWnd(pWnd);
 	if (!OpenSourceManager(pWnd)) {
		TRACE_ERROR(TWERR_OPEN_DSM);
		//"Unable to load & open TWAIN Source Manager");
    } else {

		// I will settle for the system default.  Shouldn't I get a highlight
		// on system default without this call?
		SM(DG_CONTROL, DAT_IDENTITY, MSG_GETDEFAULT, &NewSourceId);
		// now do the real thing
    	bSuccess = SM(DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, &NewSourceId);
	}

    DropToState(nStartState, pWnd);
	return bSuccess;
}

void CTwain::ModalAcquire(CWnd* pWnd)
{
	pWnd = DefWnd(pWnd);
	if (BeginAcquire(pWnd)) {
		pWnd->EnableWindow(FALSE);
		// source is enabled, wait for transfer or source closed
		ModalEventLoop();
		pWnd->EnableWindow(TRUE);
		DropToState(nStartState);
	} else {
		// BeginAcquire puts everything back when it fails
	}
}


int CTwain::BeginAcquire(CWnd* pWnd)
{
	nStartState = State();
	pWnd = DefWnd(pWnd);

	if (State() >= SOURCE_MANAGER_OPEN || OpenSourceManager(pWnd)) {
		if (State() >= SOURCE_OPEN || OpenDefaultSource()) {
			if (State() >= SOURCE_ENABLED || EnableSource(pWnd)) {
				return TRUE;
			} else {
				TRACE_ERROR(TWERR_ENABLE_SOURCE);
			}
		} else {
			TRACE_ERROR(TWERR_OPEN_SOURCE);
		}
	} else {
		TRACE_ERROR(TWERR_OPEN_DSM);
	}
	DropToState(nStartState);
	return FALSE;
} // BeginAcquire


int CTwain::LoadSourceManager(void)
{
	char		szSM[256];
	OFSTRUCT	of;

	if (nState >= SOURCE_MANAGER_LOADED) {
		return TRUE;			// SM already loaded
	}

	GetWindowsDirectory(szSM, sizeof(szSM));
	if (szSM[lstrlen(szSM)-1] != '\\') {
		lstrcat(szSM, "\\");
	}
	lstrcat(szSM, DSM_FILENAME);			// could crash!
	if (OpenFile(szSM, &of, OF_EXIST) != -1) {
		hSMLib = LoadLibrary(szSM);
	} else {
		if (bTrace) TRACE("TWAIN: LoadLibrary(%s) failed\n", szSM);
		hSMLib = NULL;
	}
	if (IsValidHandle(hSMLib)) {
		pSM_Entry = (DSMENTRYPROC) GetProcAddress(hSMLib, DSM_ENTRYPOINT);
		if (pSM_Entry) {
			iAvailable = 1;
			SetState(SOURCE_MANAGER_LOADED);
		} else {
			if (bTrace) TRACE("TWAIN: GetProcAddress() failed!!\n");
			FreeLibrary(hSMLib);
			hSMLib = NULL;
		}
	} else {
		pSM_Entry = NULL;
	}

	if (nState != SOURCE_MANAGER_LOADED && bTrace) TRACE("TWAIN: LoadSourceManager() failed.\n");
	return (nState >= SOURCE_MANAGER_LOADED);
} // LoadSourceManager


int CTwain::OpenSourceManager(CWnd* pWnd)
{
	TW_INT32 hwnd32 = (TW_INT32)(int)(DefWnd(pWnd)->m_hWnd);

	if (LoadSourceManager()) {
		SM(DG_CONTROL, DAT_PARENT, MSG_OPENDSM, &hwnd32);
		if (nState != SOURCE_MANAGER_OPEN && bTrace) TRACE("TWAIN: OPENDSM failed.\n");
	}
	return (nState >= SOURCE_MANAGER_OPEN);
}

int CTwain::OpenDefaultSource(void)
{
	if (nState != SOURCE_MANAGER_OPEN) return FALSE;

	// open the system default source
	SourceId.ProductName[0] = '\0';
	SourceId.Id = 0;
	SM(DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, &SourceId);
	if (nState != SOURCE_OPEN && bTrace) TRACE("TWAIN: OPENDS failed.\n");
	return (nState == SOURCE_OPEN);
} // OpenDefaultSource

void CTwain::SetShowUI(BOOL bShow)
// Set flag for whether source should be enabled with
// user interface visible (bShow == TRUE) or not.
// At construction, ShowUI is set TRUE.
{
	bShowUI = bShow;
} // SetShowUI


BOOL CTwain::GetShowUI(void)
// Return state of ShowUI flag.
{
	return bShowUI;
} // GetShowUI


int CTwain::EnableSource(CWnd* pWnd)
{
	pWnd = DefWnd(pWnd);
	if (nState != SOURCE_OPEN) {
		if (bTrace) TRACE("TWAIN Warning: wrong state for EnableSource()\n"); 
		return FALSE;
	}

	twUI.ShowUI = bShowUI;
	twUI.hParent = (TW_HANDLE)(pWnd->m_hWnd);
	DS(DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, &twUI);
	if (nState != SOURCE_ENABLED && bTrace) TRACE("TWAIN: ENABLEDS failed.\n");
	// note, source will set twUI.ModalUI.
	return (nState == SOURCE_ENABLED);
} // TWAIN_EnableSource


int CTwain::DisableSource(void)
{
	if (nState == SOURCE_ENABLED) {
		DS(DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, &twUI);
		if (nState == SOURCE_ENABLED && bTrace) TRACE("TWAIN: DISABLEDS failed.\n");
	}
	return (nState < SOURCE_ENABLED);
} // DisableSource


int CTwain::CloseSource(void)
{
	rc = TWRC_SUCCESS;

	if (nState == SOURCE_ENABLED) {
		DisableSource();
	}
	if (nState == SOURCE_OPEN) {
		DS(DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, &SourceId);
		if (nState == SOURCE_OPEN && bTrace) TRACE("TWAIN: CLOSEDS failed.\n");
	}
	return (nState < SOURCE_OPEN);
} // CloseSource


int CTwain::CloseSourceManager(CWnd* pWnd)
{
	CloseSource();			// close source if open

	if (nState >= SOURCE_MANAGER_OPEN) {
		TW_INT32 hwnd32 = (TW_INT32)(int)(DefWnd(pWnd)->m_hWnd);
		SM(DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, &hwnd32);
		if (nState >= SOURCE_MANAGER_OPEN && bTrace) TRACE("TWAIN: CLOSEDSM failed.\n");
	}
	return (nState < SOURCE_MANAGER_OPEN);
}


int CTwain::UnloadSourceManager(void)
{
	rc = TWRC_SUCCESS;

	if (nState == SOURCE_MANAGER_LOADED) {
		if (hSMLib) {
			FreeLibrary(hSMLib);
			hSMLib = NULL;
		}
		pSM_Entry = NULL;
		SetState(PRE_SESSION);
	}
	return (nState == PRE_SESSION);
} // UnloadSourceManager


int CTwain::EndXfer(void)
{
	if (nState == TRANSFERRING) {
		DS(DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER, &pendingXfers);
		if (nState == TRANSFERRING && bTrace) TRACE("TWAIN: ENDXFER failed.\n");
	} else {
		if (bTrace) TRACE("TWAIN Warning: wrong state for EndXfer()\n"); 
	}
	return nState < TRANSFERRING;
} // EndXfer

int CTwain::CancelXfers(void)
{
	EndXfer();			// if transferring, cancel it

	if (nState == TRANSFER_READY) {
		DS(DG_CONTROL, DAT_PENDINGXFERS, MSG_RESET, &pendingXfers);
		if (nState == TRANSFER_READY && bTrace) TRACE("TWAIN: PENDINGXFERS,RESET failed.\n");
	}
	return (nState < TRANSFER_READY);
} // CancelXfers


int CTwain::DropToState(int nS, CWnd* pWnd)
{
	pWnd = DefWnd(pWnd);
	while (nState > nS) {
		switch (nState) {
			case TRANSFERRING:
				if (!EndXfer()) return FALSE;
				break;
			case TRANSFER_READY:
				if (!CancelXfers()) return FALSE;
				break;
			case SOURCE_ENABLED:
				if (!DisableSource()) return FALSE;
				break;
			case SOURCE_OPEN:
				if (!CloseSource()) return FALSE;
				break;
			case SOURCE_MANAGER_OPEN:
				if (!CloseSourceManager(pWnd)) return FALSE;
				break;
			case SOURCE_MANAGER_LOADED:
				if (!UnloadSourceManager()) return FALSE;
				break;
			default:
				ASSERT(FALSE);
		} // switch
	} // while
	return TRUE;
}


void CTwain::ModalEventLoop(void)
{
	MSG msg;

	while ((nState >= SOURCE_ENABLED) && GetMessage((LPMSG)&msg, NULL, 0, 0)) {
		if (!TwainMessageHook ((LPMSG)&msg)) {
			TranslateMessage ((LPMSG)&msg);
			DispatchMessage ((LPMSG)&msg);
		}
	} // while
} // ModalEventLoop


int CTwain::TwainMessageHook(LPMSG lpmsg)
// returns TRUE if message processed by TWAIN
// FALSE otherwise
{
	int		bProcessed = FALSE;

	if (nState >= SOURCE_ENABLED) {
		// source enabled
		TW_EVENT	twEvent;
		twEvent.pEvent = (TW_MEMREF)lpmsg;
		twEvent.TWMessage = MSG_NULL;
		// relay message to source in case it wants it
		DS(DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, &twEvent);
		bProcessed = (rc == TWRC_DSEVENT);
		switch (twEvent.TWMessage) {
			case MSG_XFERREADY:
				// notify by callback
				// default callback does transfers
				XferReady(lpmsg);
				break;

			case MSG_CLOSEDSREQ:
				// notify by callback
				// default callback closes the source
				CloseDsRequest();
				break;

			case MSG_NULL:
				// no message returned from DS
				bProcessed=FALSE;
				break;
		} // switch
	}

	return bProcessed;
} // TwainMessageHook



void CTwain::XferReady(LPMSG lpmsg)
// Default transfer-ready handler
// Assumes xfer mech is Native, gets the next
// DIB and calls back to DibReceived();
{
	TW_UINT32		hNative;

	ASSERT(nState == TRANSFER_READY);

	do {
		// get xfers until there are no more, or 
		DS(DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, &hNative);

		if (TWRC_XFERDONE != rc) {
			// transfer failed for some reason or another
			hNative = NULL;
			if (bTrace) TRACE("TWAIN: NATIVEXFER,GET failed.\n");
		}
		ASSERT(nState >= TRANSFER_READY);
		EndXfer();				// acknowledge & end transfer
		ASSERT(nState == TRANSFER_READY || nState == SOURCE_ENABLED);
		if (hNative) {
			// call back with DIB
			if (bTrace) TRACE0("TWAIN: calling DibReceived...\n");
			DibReceived((HGLOBAL)hNative);
		}
	} while (nState == TRANSFER_READY);
}

void CTwain::CloseDsRequest(void)
{
	CloseSource();
}


void CTwain::StateChange(int nState)
{
}

void CTwain::TwainError(TW_ERR e)
{
}


void CTwain::DibReceived(HGLOBAL hDib)
{
	GlobalFree(hDib);
}

///////////////////////////////////////////////////////////////////////
// Primitive functions
int CTwain::DS(unsigned long dg, unsigned dat, unsigned msg, void FAR *pd)
// Call the current source with a triplet
{
	int bSuccess = FALSE;
	ASSERT(nState >= SOURCE_OPEN);
	if (pSM_Entry) {
		rc = (*pSM_Entry)(&AppId, &SourceId, dg, dat, msg, (TW_MEMREF)pd);
		bSuccess = (TWRC_SUCCESS == rc);
		// model state changes!
		if (DG_CONTROL == dg) {
			if (DAT_EVENT == dat) {
				if (MSG_PROCESSEVENT == msg) {
					if (MSG_XFERREADY == ((TW_EVENT FAR *)pd)->TWMessage) {
						if (bTrace) TRACE0("TWAIN: received MSG_XFERREADY\n");
						SetState(TRANSFER_READY);
					} else if (MSG_CLOSEDSREQ == ((TW_EVENT FAR *)pd)->TWMessage) {
						// well, see TWAIN v1.5 p7-27!
						if (bTrace) TRACE0("TWAIN: received MSG_CLOSEDSREQ\n");
						SetState(SOURCE_ENABLED);
					}
				}
			} else if (DAT_PENDINGXFERS == dat) {
				if (MSG_RESET == msg) {
					if (bSuccess) SetState(SOURCE_ENABLED);
				} else if (MSG_ENDXFER == msg) {
					if (bSuccess) {
						SetState(((TW_PENDINGXFERS FAR *)pd)->Count ? TRANSFER_READY : SOURCE_ENABLED);
					}
				}
			} else if (DAT_USERINTERFACE == dat) {
				if (MSG_DISABLEDS == msg) {
					if (bSuccess) SetState(SOURCE_OPEN);
				} else if (MSG_ENABLEDS == msg) {
					if (bSuccess) SetState(SOURCE_ENABLED);
				}
			} else if (DAT_IDENTITY == dat) {
				if (MSG_CLOSEDS == msg) {
					if (bSuccess) SetState(SOURCE_MANAGER_OPEN);
				}
			}
		} else if (DG_IMAGE == dg) {
			if (DAT_IMAGENATIVEXFER == dat) {
				if (MSG_GET == msg) {
					if (TWRC_XFERDONE == rc) {
						// xfer successful, still in State 7
						SetState(TRANSFERRING);
					} else if (TWRC_CANCEL == rc) {
						// xfer cancelled by user, still in State 7
						SetState(TRANSFERRING);
					} else {
						// something else happened
						// transition to State 7 failed
						// STAY IN STATE 6
						SetState(TRANSFER_READY);
					}
				}
			}
		}
	}
	return bSuccess;
} // DS



int CTwain::SM(unsigned long dg, unsigned dat, unsigned msg, void FAR *pd)
// Call the Source Manager with a triplet
{
	int bSuccess = FALSE;
	ASSERT(nState >= SOURCE_MANAGER_LOADED);
	ASSERT(pSM_Entry);
	if (pSM_Entry) {
		rc = (*pSM_Entry)(&AppId, NULL, dg, dat, msg, (TW_MEMREF)pd);
		bSuccess = (TWRC_SUCCESS == rc);
		if (DG_CONTROL == dg) {
			if (DAT_PARENT == dat) {
				if (MSG_OPENDSM == msg) {
					if (bSuccess) SetState(SOURCE_MANAGER_OPEN);
				} else if (MSG_CLOSEDSM == msg) {
					if (bSuccess) SetState(SOURCE_MANAGER_LOADED);
				}
			} else if (DAT_IDENTITY == dat) {
				if (MSG_OPENDS == msg) {
					if (bSuccess) SetState(SOURCE_OPEN);
				}
			}
		}
	}
	return bSuccess;
} // SM



unsigned CTwain::ResultCode(void)
{
	return rc;
} // ResultCode



unsigned CTwain::ConditionCode(void)
{
	TW_STATUS	twStatus;

	if (nState >= SOURCE_OPEN) {
		// get source status if open
		DS(DG_CONTROL, DAT_STATUS, MSG_GET, (TW_MEMREF)&twStatus);
	} else if (nState == SOURCE_MANAGER_OPEN) {
		// otherwise get source manager status
		SM(DG_CONTROL, DAT_STATUS, MSG_GET, (TW_MEMREF)&twStatus);
	} else {
		// nothing open, not a good time to get condition code!
		return TWCC_SEQERROR;
	}
	if (rc == TWRC_SUCCESS) {
		return twStatus.ConditionCode;
	} else {
		return TWCC_BUMMER;			// what can I say. 
	}
} // ConditionCode


